# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#

"""
this module has some handler utilities for HTTP request and response
"""
import collections
import urllib
import json
from functools import partial

from tornado.httpclient import AsyncHTTPClient, HTTPClient, HTTPRequest
from tornado.gen import coroutine, Return

JSON_TYPE = 'application/json'
DEFAULT_HEADERS = (('Content-Type', 'application/json'),)


def convert(data):
    """
    convert a standalone unicode string or unicode strings in a
    mapping or iterable into byte strings.
    """
    if isinstance(data, unicode):
        return data.encode('utf-8')
    elif isinstance(data, str):
        return data
    elif isinstance(data, collections.Mapping):
        return dict(map(convert, data.iteritems()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(convert, data))
    else:
        return data


def make_request(request, method, default_headers=None, **kwargs):
    """
    convert parameters into relevant parts
    of the an http request
    :param request: either url or an HTTPRequest object
    :param method: http request method
    :param default_headers: default headers
    :param kwargs: headers, query string params for GET and DELETE,
    data for POST and PUT
    """
    kwargs = convert(kwargs)
    if not default_headers:
        headers = dict(DEFAULT_HEADERS)
    else:
        headers = default_headers.copy()
    if isinstance(request, HTTPRequest):
        headers.update(request.headers)
    if 'headers' in kwargs:
        headers.update(kwargs.pop('headers'))
    if isinstance(request, HTTPRequest):
        request.method = method
        request.headers.update(headers)
    else:
        request = HTTPRequest(request, method, headers)
    if kwargs:
        if method in ['GET', 'DELETE']:
            request.url = "{}?{}".format(request.url, urllib.urlencode(kwargs))
        elif method in ['POST', 'PUT']:
            if request.headers['Content-Type'] == JSON_TYPE:
                request.body = json.dumps(kwargs)
            elif 'body' in kwargs:
                request.body = kwargs['body']
    return request


class ResponseObject(dict):
    """
    Access the value of in a dictionary by using
    the key as attribute.
    >>> obj = ResponseObject()
    >>> obj['foo'] = 'bar'
    >>> obj.foo
    'bar'
    """
    def __getattr__(self, key):
        return self.__getitem__(key)


def parse_response(response):
    """
    parse response and return a dictionary if the content type.
    is json/application.
    :param response: HTTPRequest
    :return dictionary for json content type otherwise response body
    """
    if response.headers.get('Content-Type', JSON_TYPE).startswith(JSON_TYPE):
        return ResponseObject(json.loads(response.body))
    else:
        return response.body


def sync_fetch(request, method, default_headers=None,
               httpclient=None, **kwargs):
    """
    fetch resource using the synchronous HTTPClient
    :param request: HTTPRequest object or a url
    :param method: HTTP method in string format, e.g. GET, POST
    :param kwargs: query string entities or POST data
    """
    updated_request = make_request(request, method, default_headers, **kwargs)
    if not httpclient:
        httpclient = HTTPClient()
    rsp = httpclient.fetch(updated_request)
    return parse_response(rsp)


@coroutine
def async_fetch(request, method, default_headers=None,
                callback=None, httpclient=None, **kwargs):
    """
    fetch resource using the asynchronous AsyncHTTPClient
    :param request: HTTPRequest object or a url
    :param method: HTTP method in string format, e.g. GET, POST
    :param callback: callback function on the result. it is used
    by the coroutine decorator.
    :param kwargs: query string entities or POST data
    """
    updated_request = make_request(request, method, default_headers, **kwargs)
    if not httpclient:
        httpclient = AsyncHTTPClient()
    rsp = yield httpclient.fetch(updated_request)
    raise Return(parse_response(rsp))


def make_fetch_func(base_url, async, **kwargs):
    """
    make a fetch function based on conditions of
    1) async
    2) ssl
    """
    if async:
        client = AsyncHTTPClient(force_instance=True, defaults=kwargs)
        return partial(async_fetch, httpclient=client)
    else:
        client = HTTPClient(force_instance=True, defaults=kwargs)
        return partial(sync_fetch, httpclient=client)



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

import logging
import calendar
import urllib
from datetime import datetime

from tornado.gen import coroutine, Return

from . import API

CLIENT_CREDENTIALS = 'client_credentials'
JWT_BEARER = 'urn:ietf:params:oauth:grant-type:jwt-bearer'


class Read(object):
    """Object representing a read scope"""
    def __init__(self, resource_id=None):
        """
        Scope for reading a resource or all resources

        :param resource_id: (optional) the resource ID e.g. service or
            repository ID. If resource_id is not provided the scope will be
            "read"
        """
        self.resource_id = resource_id

    def __repr__(self):
        return '<{}>'.format(str(self))

    def __str__(self):
        if self.resource_id is None:
            return 'read'
        else:
            return 'read[{}]'.format(self.resource_id)

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def __hash__(self):
        return hash(str(self))


class Write(object):
    """Object representing a write scope"""
    def __init__(self, resource_id):
        """
        Scope for writing to a resource

        :param resource_id: the resource ID e.g. service or repository ID
        """
        self.resource_id = resource_id

    def __repr__(self):
        return '<{}>'.format(str(self))

    def __str__(self):
        return 'write[{}]'.format(self.resource_id)

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def __hash__(self):
        return hash(str(self))


class Delegate(object):
    """Object representing a delegate scope"""
    def __init__(self, resource_id, access):
        """
        Scope for delegating access to a resource

        :param resource_id: the resource ID e.g. service or repository ID
        :param access: a Read or Write object, or an appropriate string (e.g.
            "write:1234")
        """
        self.resource_id = resource_id
        self.access = access

    def __repr__(self):
        return '<{}>'.format(str(self))

    def __str__(self):
        return 'delegate[{}]:{}'.format(self.resource_id, self.access)

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def __hash__(self):
        return hash(str(self))


class Scope(object):
    def __init__(self, *scopes):
        """
        A scope object representing a space separated list of scopes

        :param scopes: Read, Write or Delegate objects
        """
        self._scopes = set(scopes)

    def add(self, scope):
        """Add a scope"""
        self._scopes.add(scope)

    def remove(self, scope):
        """Remove a scope """
        try:
            self._scopes.remove(scope)
        except KeyError:
            pass

    def __repr__(self):
        return '<Scope "{}">'.format(str(self))

    def __str__(self):
        return ' '.join(map(str, sorted(self._scopes)))

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)


class RequestToken(object):
    """
    Function object for making token requests with simple caching of the
    responses

    NOTE: errors are not cached
    """
    # Don't re-use a token with less than this many seconds remaining
    max_until_expired = 60
    max_cache_size = 100

    def __init__(self):
        self._cache = {}

    @coroutine
    def __call__(self, base_url, client_id, client_secret, scope=None,
                 jwt=None, cache=True, **kwargs):
        """
        Get an OAuth token

        Example Usage:
            >>> token = get_token('https://localhost:8007',
                                'a service id',
                                'the client secret')

        With a scope:
            >>> token = get_token('https://localhost:8007',
                                'a service id',
                                'the client secret',
                                scope=Scope(Write('another service'),
                                            Write('yet another service')))

            >>> token = get_token('https://localhost:8007',
                                'a service id',
                                'the client secret',
                                scope=Write('just one service'))

        Scope can also just be a string:
            >>> token = get_token('https://localhost:8007',
                                'a service id',
                                'the client secret',
                                scope='delegate:1234:write:5678')

        Exchange a delegate token:
            >>> delegate_token = get_token('https://localhost:8007',
                                        'a service id',
                                        'the client secret',
                                        scope='delegate:1234:write:5678')
            >>> token = get_token('https://localhost:8007',
                                '1234',
                                '1234 secret',
                                jwt=delegate_token,
                                scope='write:5678')


        :param base_url: base auth serivce URL
        :param client_id: the client ID
        :param client_secret: the client secret
        :param scope: (optional) a scope object or string
        :param jwt: (optional) a JWT for the jwt-bearer grant
        :param cache: whether or not to use the in-memory cache
        :param ca_certs: (optional) a CA certificate file
        """
        parameters = {'grant_type': CLIENT_CREDENTIALS}

        if scope is not None:
            parameters['scope'] = str(scope)

        if jwt is not None:
            parameters['grant_type'] = JWT_BEARER
            parameters['assertion'] = jwt

        if cache:
            response = yield self._cached_request(base_url, client_id,
                                                  client_secret, parameters,
                                                  **kwargs)
        else:
            response = yield self._request(base_url, client_id,
                                           client_secret, parameters,
                                           **kwargs)

        raise Return(response['access_token'])

    @coroutine
    def _request(self, base_url, client_id, client_secret,
                 parameters, **kwargs):
        """Make an API request to get the token"""
        logging.debug('Getting an OAuth token for client "%s" with scope "%s"',
                      client_id, parameters.get('scope'))
        headers = {'Content-Type': 'application/x-www-form-urlencoded',
                   'Accept': 'application/json'}

        api = API(base_url,
                  auth_username=client_id,
                  auth_password=client_secret,
                  **kwargs)
        endpoint = api.auth.token

        response = yield endpoint.post(body=urllib.urlencode(parameters),
                                       request_timeout=60,
                                       headers=headers)

        logging.debug('Received token: %s', response.get('access_token'))
        raise Return(response)

    @staticmethod
    def _now():
        return calendar.timegm(datetime.utcnow().timetuple())

    def _expired(self, cached_item):
        max_expiry = self._now() + self.max_until_expired
        return cached_item.get('expiry') < max_expiry

    @coroutine
    def _cached_request(self, base_url, client_id, client_secret,
                        parameters, **kwargs):
        """Cache the token request and use cached responses if available"""
        key = (base_url, client_id, tuple(parameters.items()))
        cached = self._cache.get(key, {})

        if not cached.get('access_token') or self._expired(cached):
            cached = yield self._request(base_url, client_id, client_secret,
                                         parameters, **kwargs)
            self._cache[key] = cached
            # Purge cache when adding a new item so it doesn't grow too large
            # It's assumed the cache size is small enough that it's OK to loop
            # over the whole cache regularly. If not, could change this to
            # just pop off the oldest one
            self.purge_cache()

        logging.debug('Using a cached token: %s', cached.get('access_token'))
        raise Return(cached)

    def reset_cache(self):
        """Reset the token cache"""
        self._cache = {}

    def purge_cache(self):
        """
        Purge expired cached tokens and oldest tokens if more than cache_size
        """
        if len(self._cache) > self.max_cache_size:
            items = sorted(self._cache.items(), key=lambda (k, v): v['expiry'])
            self._cache = {k: v for k, v in items[self.max_cache_size:]
                           if not self._expired(v)}

get_token = RequestToken()



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#

from setuptools import setup
import re

with open('chub/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

setup(name='opp-chub',
      version=version,
      description='Asynchronous Python client for the Open Permissions Platform Coalition REST Services',
      author='Open Permissions Platform Coalition',
      author_email='support@openpermissions.org',
      url='https://github.com/openpermissions/chub',
      packages=['chub'],
      install_requires=['tornado'],
      license='Apache 2.0',
      classifiers=(
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Natural Language :: English',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: Implementation :: CPython',
      ),
      include_package_data=True,
      package_data={
            '': [ 'LICENSE' ]
      }
      )



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

"""
this module has classes for converting python method calls
to request on RESTful api
"""
import inspect
from functools import partial
from urlparse import urljoin
from urllib import quote_plus

from tornado.httpclient import HTTPRequest

from .handlers import make_fetch_func, DEFAULT_HEADERS


HTTP_METHODS = ['GET', 'POST', 'HEAD', 'PUT', 'PATCH', 'DELETE',
                'OPTIONS', 'TRACE', 'CONNECT']

API_VERSION = 'v1'


class Resource(object):
    """
    Resource converts python method calls to request on RESTful api
    """

    def __init__(self, path, fetch, resource_map=None,
                 request_class=HTTPRequest, default_headers=None):
        self.path = path
        self.fetch = fetch
        if resource_map is None:
            self.resource_map = {}
        else:
            self.resource_map = resource_map
        self.request_class = request_class
        self.http_request = None
        if not default_headers:
            default_headers = dict(DEFAULT_HEADERS)
        self.default_headers = default_headers

    def __getattr__(self, key):
        if key.upper() in HTTP_METHODS:
            # tornado will take either an HTTPRequest object or a url
            # an HTTPRequest object will be used directly
            # a url will be converted into one by tornado
            request = self.http_request if self.http_request else self.path
            return partial(self.fetch, request=request, method=key.upper(),
                           default_headers=self.default_headers)
        path = '/'.join((self.path, key))
        return self._sub_resource(path)

    def _sub_resource(self, path):
        """
        get or create sub resource
        """
        if path not in self.resource_map:
            self.resource_map[path] = Resource(
                path, self.fetch, self.resource_map,
                default_headers=self.default_headers)
        return self.resource_map[path]

    def __getitem__(self, entity_id):
        path = '/'.join((self.path, quote_plus(entity_id)))
        return self._sub_resource(path)

    def prepare_request(self, *args, **kw):
        """
        creates a full featured HTTPRequest objects
        """
        self.http_request = self.request_class(self.path, *args, **kw)


class API(Resource):
    """
    API takes the base_url and a boolean async . Based on value of async
    an async or sync fetch function is set
    """

    mappings = {}

    def __init__(self, base_url, async=True, api_version=API_VERSION,
                 token=None, **kwargs):
        self.base_url = urljoin(base_url, api_version)
        fetch = make_fetch_func(self.base_url, async, **kwargs)
        super(API, self).__init__(self.base_url, fetch)
        if token:
            self.token = token

    @property
    def token(self):
        """
        get the token
        """
        header = self.default_headers.get('Authorization', '')
        prefex = 'Bearer '
        if header.startswith(prefex):
            token = header[len(prefex):]
        else:
            token = header
        return token

    @token.setter
    def token(self, token):
        """
        set the token
        """
        self.default_headers['Authorization'] = 'Bearer {}'.format(token)

    def _request(self):
        """
        retrieve the caller frame, extract the parameters from
        the caller function, find the matching function,
        and fire the request
        """
        caller_frame = inspect.getouterframes(inspect.currentframe())[1]
        args, _, _, values = inspect.getargvalues(caller_frame[0])
        caller_name = caller_frame[3]
        kwargs = {arg: values[arg] for arg in args if arg != 'self'}
        func = reduce(
            lambda resource, name: resource.__getattr__(name),
            self.mappings[caller_name].split('.'), self)
        return func(**kwargs)


class Accounts(API):
    """
    API client for accounts service
    """

    mappings = {
        'info': 'accounts.get',
        'login': 'accounts.login.post',
        'register_service': 'accounts.services.post'}

    def info(self):
        """
        get service information
        """
        return self._request()

    def login(self, email, password):
        """
        login using email and password
        :param email: email address
        :param password: password
        """
        rsp = self._request()
        self.default_headers['Authorization'] = rsp.data['token']
        return rsp

    def register_service(self, name, location, organisation_id, certificate):
        """
        register a service
        :param name: name of the service
        :param location: url of the service
        :param organisation_id: organisation id
        :param certificate: ssl certificate
        """
        return self._request()
